package cn.com.qmth.examcloud.web.facepp;

import java.util.concurrent.TimeUnit;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.compress.utils.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.http.HttpEntity;
import org.apache.http.HttpStatus;
import org.apache.http.client.config.CookieSpecs;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.mime.MultipartEntityBuilder;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.apache.http.util.EntityUtils;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;

import cn.com.qmth.examcloud.commons.exception.ExamCloudRuntimeException;
import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.helpers.JsonHttpResponseHolder;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLog;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLogFactory;
import cn.com.qmth.examcloud.web.bootstrap.PropertyHolder;

/**
 * face++ 客户端
 *
 * @author WANGWEI
 * @date 2019年9月16日
 * @Copyright (c) 2018-? http://qmth.com.cn All Rights Reserved.
 */
public class FaceppClient {

	protected static ExamCloudLog log = ExamCloudLogFactory.getLog(FaceppClient.class);

	private static CloseableHttpClient httpclient;

	private static RequestConfig requestConfig;

	private static FaceppClient faceppClient;

	private static String apiKey;

	private static String apiSecret;

	private static String compareUrl = "https://api-cn.faceplusplus.com/facepp/v3/compare";

	private FaceppClient() {
	}

	/**
	 * 获取单例
	 *
	 * @author WANGWEI
	 * @return
	 */
	public static FaceppClient getClient() {
		if (null == faceppClient) {
			synchronized (FaceppClient.class) {
				if (null == faceppClient) {
					faceppClient = new FaceppClient();
				}
			}
		}

		return faceppClient;
	}

	static {
		PoolingHttpClientConnectionManager cm = new PoolingHttpClientConnectionManager(60,
				TimeUnit.SECONDS);
		cm.setValidateAfterInactivity(1000);
		cm.setMaxTotal(8000);
		cm.setDefaultMaxPerRoute(200);

		requestConfig = RequestConfig.custom().setConnectionRequestTimeout(500)
				.setSocketTimeout(10000).setConnectTimeout(10000)
				.setCookieSpec(CookieSpecs.IGNORE_COOKIES).build();

		httpclient = HttpClients.custom().setConnectionManager(cm).disableAutomaticRetries()
				.setDefaultRequestConfig(requestConfig).build();

		apiKey = PropertyHolder.getString("$facepp.apiKey");
		apiSecret = PropertyHolder.getString("$facepp.apiSecret");

		if (StringUtils.isBlank(apiKey)) {
			log.error("'facepp.apiKey' is not configured");
		}
		if (StringUtils.isBlank(apiSecret)) {
			log.error("'facepp.apiSecret' is not configured");
		}
	}

	/**
	 * 人脸识别
	 *
	 * @author WANGWEI
	 * @param faceToken
	 * @param imageUrl
	 * @return
	 */
	public JsonHttpResponseHolder compareWithTokenAndImageUrl(String faceToken, String imageUrl) {

		if (log.isDebugEnabled()) {
			log.debug("[Face++]. compareWithTokenAndImageUrl(String,String); faceToken=" + faceToken
					+ "; imageUrl=" + imageUrl);
		}

		String url = System.getProperty("facepp.compare.url");
		if (StringUtils.isBlank(url)) {
			url = compareUrl;
		}

		HttpPost httpPost = new HttpPost(url);
		httpPost.setConfig(FaceppClient.requestConfig);

		MultipartEntityBuilder builder = MultipartEntityBuilder.create();
		builder.addTextBody("api_key", apiKey);
		builder.addTextBody("api_secret", apiSecret);
		builder.addTextBody("face_token1", faceToken);
		builder.addTextBody("image_url2", imageUrl);
		HttpEntity httpEntity = builder.build();

		httpPost.setEntity(httpEntity);

		CloseableHttpResponse response = null;
		JsonHttpResponseHolder responseHolder = null;
		long s = System.currentTimeMillis();
		try {

			response = httpclient.execute(httpPost);
			int statusCode = response.getStatusLine().getStatusCode();
			String entityStr = EntityUtils.toString(response.getEntity(), "UTF-8");
			JSONObject obj = JSON.parseObject(entityStr);
			responseHolder = new JsonHttpResponseHolder(statusCode, obj);

			if (HttpStatus.SC_OK != responseHolder.getStatusCode()) {
				log.error("[Face++]. compareWithTokenAndImageUrl(String,String); statusCode="
						+ statusCode + "; responseEntity=" + entityStr);
			} else {
				if (log.isDebugEnabled()) {
					log.debug("[Face++]. compareWithTokenAndImageUrl(String,String); statusCode="
							+ statusCode + "; responseEntity=" + entityStr);
				}
			}

		} catch (Exception e) {
			throw new ExamCloudRuntimeException(e);
		} finally {
			IOUtils.closeQuietly(response);
		}

		if (log.isDebugEnabled()) {
			log.debug("[Face++]. compareWithTokenAndImageUrl(String,String); faceToken=" + faceToken
					+ "; imageUrl=" + imageUrl + "; cost " + (System.currentTimeMillis() - s)
					+ " ms.");
		}

		return responseHolder;
	}

	/**
	 * 人脸识别<br>
	 * 优先使用主地址调用face++<br>
	 * 主地址无效时,使用备用地址调用face++.<br>
	 * 备用地址也无效时,使用备用地址下载数据,使用下载数据的base64加密串调用face++<br>
	 *
	 * @author WANGWEI
	 * 
	 * @param faceToken
	 *            face++预存照片
	 * @param imageUrl
	 *            主地址
	 * @param backupImageUrl
	 *            备用地址
	 * @return
	 * @throws StatusException
	 *             code为801,802,803表示图片地址无效
	 */
	public JsonHttpResponseHolder compareWithTokenAndImageUrl(String faceToken, String imageUrl,
			String backupImageUrl) throws StatusException {

		JsonHttpResponseHolder responseHolder = compareWithTokenAndImageUrl(faceToken, imageUrl);

		if (HttpStatus.SC_OK == responseHolder.getStatusCode()) {
			return responseHolder;
		}

		JSONObject respBody = responseHolder.getRespBody();
		String errMsg = respBody.getString("error_message");
		if (retry(errMsg)) {
			responseHolder = compareWithTokenAndImageUrl(faceToken, backupImageUrl);
		}

		respBody = responseHolder.getRespBody();
		errMsg = respBody.getString("error_message");

		if (retry(errMsg)) {
			HttpGet get = new HttpGet(backupImageUrl);
			get.setConfig(FaceppClient.requestConfig);
			CloseableHttpResponse response = null;
			String imageBase64 = null;
			long s = System.currentTimeMillis();
			try {
				response = httpclient.execute(get);

				if (HttpStatus.SC_OK != response.getStatusLine().getStatusCode()) {
					throw new StatusException("801",
							"fail to download file. url=" + backupImageUrl);
				}

				byte[] byteArray = EntityUtils.toByteArray(response.getEntity());
				if (100 > byteArray.length) {
					throw new StatusException("802", "invalid image size. url=" + backupImageUrl);
				}

				imageBase64 = Base64.encodeBase64String(byteArray);

			} catch (StatusException e) {
				log.error("fail to download file. url=" + backupImageUrl, e);
				throw e;
			} catch (Exception e) {
				log.error("fail to download file. url=" + backupImageUrl, e);
				throw new StatusException("803", "fail to download file. url=" + backupImageUrl, e);
			} finally {
				IOUtils.closeQuietly(response);
			}

			if (log.isDebugEnabled()) {
				log.debug(
						"[Face++]. compareWithTokenAndImageUrl(String,String,String); download image; url="
								+ backupImageUrl + "; cost " + (System.currentTimeMillis() - s)
								+ " ms.");
			}

			responseHolder = compareWithTokenAndBase64(faceToken, imageBase64);
		}

		return responseHolder;
	}

	private boolean retry(String errMsg) {
		if (null != errMsg) {
			if (errMsg.startsWith("INVALID_IMAGE_URL")
					|| errMsg.startsWith("IMAGE_DOWNLOAD_TIMEOUT")) {
				return true;
			}
		}
		return false;
	}

	/**
	 * 人脸识别
	 *
	 * @author WANGWEI
	 * @param faceToken
	 * @param imageBase64
	 * @return
	 */
	public JsonHttpResponseHolder compareWithTokenAndBase64(String faceToken, String imageBase64) {

		String url = System.getProperty("facepp.compare.url");
		if (StringUtils.isBlank(url)) {
			url = compareUrl;
		}

		HttpPost httpPost = new HttpPost(url);
		httpPost.setConfig(FaceppClient.requestConfig);

		MultipartEntityBuilder builder = MultipartEntityBuilder.create();
		builder.addTextBody("api_key", apiKey);
		builder.addTextBody("api_secret", apiSecret);
		builder.addTextBody("face_token1", faceToken);
		builder.addTextBody("image_base64_2", imageBase64);
		HttpEntity httpEntity = builder.build();

		httpPost.setEntity(httpEntity);

		CloseableHttpResponse response = null;
		JsonHttpResponseHolder responseHolder = null;
		long s = System.currentTimeMillis();
		try {

			response = httpclient.execute(httpPost);
			int statusCode = response.getStatusLine().getStatusCode();
			String entityStr = EntityUtils.toString(response.getEntity(), "UTF-8");
			JSONObject obj = JSON.parseObject(entityStr);
			responseHolder = new JsonHttpResponseHolder(statusCode, obj);

			if (HttpStatus.SC_OK != responseHolder.getStatusCode()) {
				log.error("[Face++]. compareWithTokenAndBase64(String,String); statusCode="
						+ statusCode + "; responseEntity=" + entityStr);
			} else {
				if (log.isDebugEnabled()) {
					log.debug("[Face++]. compareWithTokenAndBase64(String,String); statusCode="
							+ statusCode + "; responseEntity=" + entityStr);
				}
			}

		} catch (Exception e) {
			throw new ExamCloudRuntimeException(e);
		} finally {
			IOUtils.closeQuietly(response);
		}

		if (log.isDebugEnabled()) {
			log.debug("[Face++]. compareWithTokenAndBase64(String,String); faceToken=" + faceToken
					+ "; imageBase64=" + imageBase64 + "; cost " + (System.currentTimeMillis() - s)
					+ " ms.");
		}

		return responseHolder;
	}

	/**
	 * 人脸识别
	 *
	 * @author WANGWEI
	 * @param imageUrl1
	 * @param imageUrl2
	 * @return
	 */
	public JsonHttpResponseHolder compareWithImageUrl(String imageUrl1, String imageUrl2) {

		String url = System.getProperty("facepp.compare.url");
		if (StringUtils.isBlank(url)) {
			url = compareUrl;
		}

		HttpPost httpPost = new HttpPost(url);
		httpPost.setConfig(FaceppClient.requestConfig);

		MultipartEntityBuilder builder = MultipartEntityBuilder.create();
		builder.addTextBody("api_key", apiKey);
		builder.addTextBody("api_secret", apiSecret);
		builder.addTextBody("image_url1", imageUrl1);
		builder.addTextBody("image_url2", imageUrl2);
		HttpEntity httpEntity = builder.build();

		httpPost.setEntity(httpEntity);

		CloseableHttpResponse response = null;
		JsonHttpResponseHolder responseHolder = null;
		long s = System.currentTimeMillis();
		try {

			response = httpclient.execute(httpPost);
			int statusCode = response.getStatusLine().getStatusCode();
			String entityStr = EntityUtils.toString(response.getEntity(), "UTF-8");
			JSONObject obj = JSON.parseObject(entityStr);
			responseHolder = new JsonHttpResponseHolder(statusCode, obj);

			if (HttpStatus.SC_OK != responseHolder.getStatusCode()) {
				log.error("[Face++]. compareWithImageUrl(String,String); statusCode=" + statusCode
						+ "; responseEntity=" + entityStr);
			} else {
				if (log.isDebugEnabled()) {
					log.debug("[Face++]. compareWithImageUrl(String,String); statusCode="
							+ statusCode + "; responseEntity=" + entityStr);
				}
			}

		} catch (Exception e) {
			throw new ExamCloudRuntimeException(e);
		} finally {
			IOUtils.closeQuietly(response);
		}

		if (log.isDebugEnabled()) {
			log.debug("[Face++]. compareWithImageUrl(String,String); image_url1=" + imageUrl1
					+ "; imageUrl2=" + imageUrl2 + "; cost " + (System.currentTimeMillis() - s)
					+ " ms.");
		}

		return responseHolder;
	}

}
